/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 by the Free Software Foundation.               *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "liberty.h"

#include "ttsgf.h"
#include "sgfana.h"

#define MAXTRY 400

void sgfShowConsideredMoves(void);

/*
 * Private data structures used to collect information about moves.
 * FIXME: Should these reside elsewhere?
 * FIXME: The static allocation should be replaced by some dynamic scheme.
 *        Also the data structures in general should probably be redesigned.
 */

#define MAX_MOVE_REASONS 1000
#define MAX_STRINGS 2*MAX_BOARD*MAX_BOARD/3
#define MAX_DRAGONS MAX_STRINGS
#define MAX_CONNECTIONS 4*MAX_STRINGS

static struct move_data move[MAX_BOARD][MAX_BOARD];
static struct move_reason move_reasons[MAX_MOVE_REASONS];
static int next_reason;

/* Strings */
static int stringi[MAX_STRINGS];
static int stringj[MAX_STRINGS];
static int next_string;

/* Dragons */
static int dragoni[MAX_DRAGONS];
static int dragonj[MAX_DRAGONS];
static int next_dragon;

/* Connections */
static int conn_dragon1[MAX_CONNECTIONS];
static int conn_dragon2[MAX_CONNECTIONS];
static int next_connection;

/* Initialize move reason data structures. */
static void
clear_move_reasons(void)
{
  int i;
  int j;
  int k;
  next_reason = 0;
  next_string = 0;
  next_dragon = 0;
  next_connection = 0;

  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      move[i][j].value = 0;
      for (k=0; k<MAX_REASONS; k++)
	move[i][j].reason[k] = -1;
    }
}

/*
 * Find the index of a string in the list of strings. If necessary,
 * add a new entry. (ai, aj) must point to the origin of the string.
 */
static int
find_string(int ai, int aj)
{
  int k;
  for (k=0; k<next_string; k++)
    if ((stringi[k] == ai) && (stringj[k] == aj))
      return k;

  /* Add a new entry. */
  assert(next_string < MAX_STRINGS);
  stringi[next_string] = ai;
  stringj[next_string] = aj;
  next_string++;
  return next_string - 1;
}

/*
 * Find the index of a dragon in the list of dragons. If necessary,
 * add a new entry. (ai, aj) must point to the origin of the dragon.
 */
static int
find_dragon(int ai, int aj)
{
  int k;
  for (k=0; k<next_dragon; k++)
    if ((dragoni[k] == ai) && (dragonj[k] == aj))
      return k;

  /* Add a new entry. */
  assert(next_dragon < MAX_DRAGONS);
  dragoni[next_dragon] = ai;
  dragonj[next_dragon] = aj;
  next_dragon++;
  return next_dragon - 1;
}

/*
 * Find the index of a connection in the list of connections.
 * If necessary, add a new entry.
 */
static int
find_connection(int dragon1, int dragon2)
{
  int k;

  if (dragon1 > dragon2) {
    /* Swap to canonical order. */
    int tmp = dragon1;
    dragon1 = dragon2;
    dragon2 = tmp;
  }
  
  for (k=0; k<next_connection; k++)
    if ((conn_dragon1[k] == dragon1) && (conn_dragon2[k] == dragon2))
      return k;

  /* Add a new entry. */
  assert(next_connection < MAX_CONNECTIONS);
  conn_dragon1[next_connection] = dragon1;
  conn_dragon2[next_connection] = dragon2;
  next_connection++;
  return next_connection - 1;
}

/*
 * Find a reason in the list of reasons. If necessary, add a new entry.
 */
static int
find_reason(int type, int what)
{
  int k;
  for (k=0; k<next_reason; k++)
    if ((move_reasons[k].type == type) && (move_reasons[k].what == what))
      return k;

  /* Add a new entry. */
  assert(next_reason < MAX_MOVE_REASONS);
  move_reasons[next_reason].type = type;
  move_reasons[next_reason].what = what;
  next_reason++;
  return next_reason - 1;
}

/*
 * Add a move reason for (ti, tj) if it's not already there or the
 * table is full.
 */ 
static void
add_move_reason(int ti, int tj, int type, int what)
{
  int k;
  for (k=0; k<MAX_REASONS; k++) {
    int r = move[ti][tj].reason[k];
    if (r < 0)
      break;
    if (move_reasons[r].type == type
	&& move_reasons[r].what == what)
      return;  /* Reason already listed. */
  }
  /* Reason not found, add it if there is place left. */
  if (k<MAX_REASONS)
    move[ti][tj].reason[k] = find_reason(type, what);
}

/*
 * Add to the reasons for the move at (ti, tj) that it defends the string
 * at (ai, aj).
 */
void
add_defense_move(int ti, int tj, int ai, int aj)
{
  int string_number = find_string(worm[ai][aj].origini,
				  worm[ai][aj].originj);
  add_move_reason(ti, tj, DEFEND_MOVE, string_number);
}
  
/*
 * Add to the reasons for the move at (ti, tj) that it attacks the string
 * at (ai, aj).
 */
void
add_attack_move(int ti, int tj, int ai, int aj)
{
  int string_number = find_string(worm[ai][aj].origini, worm[ai][aj].originj);
  add_move_reason(ti, tj, ATTACK_MOVE, string_number);
}

/*
 * Remove from the reasons for the move at (ti, tj) that it defends
 * the string at (ai, aj).
 *
 * FIXME: Not implemented.
 */
void
remove_defense_move(int ti, int tj, int ai, int aj)
{
}


/*
 * Remove from the reasons for the move at (ti, tj) that it attacks
 * the string at (ai, aj).
 *
 * FIXME: Not implemented.
 */
void
remove_attack_move(int ti, int tj, int ai, int aj)
{
}

/*
 * Add to the reasons for the move at (ti, tj) that it connects the
 * dragons at (ai, aj) and (bi, bj).
 */
void
add_connection_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int dragon2 = find_dragon(dragon[bi][bj].origini, dragon[bi][bj].originj);
  int connection = find_connection(dragon1, dragon2);
  add_move_reason(ti, tj, CONNECT_MOVE, connection);
}

/*
 * Add to the reasons for the move at (ti, tj) that it cuts the
 * dragons at (ai, aj) and (bi, bj).
 */
void
add_cut_move(int ti, int tj, int ai, int aj, int bi, int bj)
{
  int dragon1 = find_dragon(dragon[ai][aj].origini, dragon[ai][aj].originj);
  int dragon2 = find_dragon(dragon[bi][bj].origini, dragon[bi][bj].originj);
  int connection = find_connection(dragon1, dragon2);

  /*
   * Ignore the cut or connection if either (ai, aj) or (bi, bj)
   * points to a tactically captured worm.
   */
  if ((worm[ai][aj].attacki != -1 && worm[ai][aj].defendi == -1)
      || (worm[bi][bj].attacki != -1 && worm[bi][bj].defendi == -1))
    return;
  
  add_move_reason(ti, tj, CUT_MOVE, connection);
}

/*
 * Add to the reasons for the move at (ti, tj) that it is an anti-suji.
 * This means that it's a locally inferior move or for some other reason
 * must *not* be played.
 */
void
add_antisuji_move(int ti, int tj)
{
  add_move_reason(ti, tj, ANTISUJI_MOVE, 0);
}

#if EXPERIMENTAL_MOVEGEN

/*
 * Any move that captures or defends a worm also connects or cuts
 * the surrounding dragons. Find these secondary move reasons.
 *
 * FIXME: There is a certain amount of optimizations that could be
 *        done here.
 */
static int adji[MAXCHAIN];
static int adjj[MAXCHAIN];
static int adjsize[MAXCHAIN];
static int adjlib[MAXCHAIN];

static void
induce_secondary_move_reasons(int color)
{
  int m;
  int n;
  int k;
  int i;
  int neighbors;
  int ai;
  int aj;
  int di = -1;
  int dj = -1;
  int best_strength = 0;
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      for (k=0; k<MAX_REASONS; k++) {

	int r = move[m][n].reason[k];
	if (r == -1)
	  break;

	if (move_reasons[r].type != ATTACK_MOVE
	    && move_reasons[r].type != DEFEND_MOVE)
	  continue;

	ai = stringi[move_reasons[r].what];
	aj = stringj[move_reasons[r].what];

	if ((   (move_reasons[r].type == ATTACK_MOVE)
		&& (p[ai][aj] == color))
	    || ((move_reasons[r].type == DEFEND_MOVE)
		&& (p[ai][aj] != color)))
	  continue; /* Only a move for the opponent. */
	
	if (worm[ai][aj].defendi == -1)
	  continue; /* No defense. */

	/*
	 * If this is a defense move and the defended string is part
	 * of a dragon that is dead or has negative vitality, we only
	 * induce a cutting move if the same move also attacks or
	 * connects something.
	 */
	if (move_reasons[r].type == DEFEND_MOVE
	    && (dragon[ai][aj].status == DEAD
		|| dragon[ai][aj].vitality < 0)) {
	  int ok = 0;
	  for (i=0; i<MAX_REASONS; i++) {
	    int rr = move[m][n].reason[k];
	    if (rr == -1)
	      break;
	    if (move_reasons[rr].type == ATTACK_MOVE
		|| move_reasons[rr].type == CONNECT_MOVE) {
	      ok = 1;
	      break;
	    }
	  }
	  if (!ok)
	    continue;
	}
	
	/* Find all dragons surrounding the worm. */
	chainlinks(ai, aj, &neighbors, adji, adjj, adjsize, adjlib);

	/*
	 * Find the strongest of the surrounding dragons and say that
	 * all other dragons are connected or cut with respect to that
	 * one.
	 *
	 * For the time being we use the vitality value of the dragon
	 * to measure strength. We probably want to change this later.
	 */
	for (i=0; i<neighbors; i++)
	  if (i == 0 || dragon[adji[i]][adjj[i]].vitality > best_strength) {
	    di = dragon[adji[i]][adjj[i]].origini;
	    dj = dragon[adji[i]][adjj[i]].originj;
	    best_strength = dragon[di][dj].vitality;
	  }

	for (i=0; i<neighbors; i++) {
	  int ei = dragon[adji[i]][adjj[i]].origini;
	  int ej = dragon[adji[i]][adjj[i]].originj;

	  if (di != ei || dj != ej) {
	    if (move_reasons[r].type == ATTACK_MOVE)
	      add_connection_move(m, n, di, dj, ei, ej);
	    else
	      add_cut_move(m, n, di, dj, ei, ej);
	  }
	}
      }
}

static void
list_move_reasons(int color)
{
  int m;
  int n;
  int k;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int dragon1 = -1;
  int dragon2 = -1;
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      for (k=0; k<MAX_REASONS; k++) {

	int r = move[m][n].reason[k];
	if (r == -1)
	  break;

	switch(move_reasons[r].type) {
	  case ATTACK_MOVE:
	  case DEFEND_MOVE:
	    ai = stringi[move_reasons[r].what];
	    aj = stringj[move_reasons[r].what];

	    if (move_reasons[r].type == ATTACK_MOVE
		&& p[ai][aj] != color)
	      gprintf("Move at %m attacks %m%s\n", m, n, ai, aj,
		      (worm[ai][aj].defendi == -1) ? " (defenseless)" : "");
	    else if ((move_reasons[r].type == DEFEND_MOVE
		      && p[ai][aj] == color))
	      gprintf("Move at %m defends %m\n", m, n, ai, aj);
	    break;
	  case CONNECT_MOVE:
	  case CUT_MOVE:
	    dragon1 = conn_dragon1[move_reasons[r].what];
	    dragon2 = conn_dragon2[move_reasons[r].what];
	    ai = dragoni[dragon1];
	    aj = dragonj[dragon1];
	    bi = dragoni[dragon2];
	    bj = dragonj[dragon2];
	    if (move_reasons[r].type == CONNECT_MOVE)
	      gprintf("Move at %m connects %m and %m\n", m, n, ai, aj, bi, bj);
	    else
	      gprintf("Move at %m cuts %m and %m\n", m, n, ai, aj, bi, bj);
	    break;
	}
      }
}

/*
 * Measure the "effective" size of a dragon. This measure is
 * completely ad hoc and not even reasonable.
 *
 * FIXME: Exchange this for something that is reasonable.
 */
static int effective_dragon_size(int ai, int aj)
{
  int dragon_size = 0;
  int i;
  int j;

  /* Find all worms making up this dragon. */
  for (i=0; i<board_size; i++)
    for (j=0; j<board_size; j++) {
      if (p[i][j] == p[ai][aj]
	  && worm[i][j].origini == i
	  && worm[i][j].originj == j
	  && dragon[i][j].origini == ai
	  && dragon[i][j].originj == aj)
	dragon_size += 2 * worm[ai][aj].size
	  + worm[ai][aj].liberties + worm[ai][aj].liberties2
	  + (worm[ai][aj].liberties3 + worm[ai][aj].liberties4)/2;
    }

  return dragon_size;
}

/*
 * Strategical value of connecting (or cutting) the dragon at (ai, aj)
 * to the dragon at (bi, bj). Notice that this function is assymetric.
 */
static int
connection_value2(int ai, int aj, int bi, int bj)
{
  int status1 = dragon[ai][aj].status;
  int status2 = dragon[bi][bj].status;

  int dragon_size = effective_dragon_size(ai, aj);
  
  if (status1 == ALIVE) {
    if (dragon[ai][aj].vitality > 10)
      return 0;
    if (dragon[bi][bj].genus > 0 || dragon[bi][bj].heyes > 0
	|| status2 != DEAD)
      return dragon_size;
    else
      return 0;
  }
  
  if (status1 == CRITICAL) {
    if (status2 == ALIVE)
      return 6 * dragon_size;
    if (dragon[bi][bj].genus > 0 || dragon[bi][bj].heyes > 0
	|| status2 != DEAD)
      return 5 * dragon_size;
    else
      return 0;
  }
    
  if (status1 == UNKNOWN) {
    if (status2 == ALIVE) {
      if (dragon[ai][aj].vitality < 0)
	return 5 * dragon_size;
      else if (dragon[ai][aj].vitality < 10)
	return 3 * dragon_size;
      else
	return 2 * dragon_size;
    }
    if (dragon[bi][bj].genus > 0 || dragon[bi][bj].heyes > 0
	|| status2 != DEAD) {
      if (dragon[ai][aj].vitality < 0)
	return 4 * dragon_size;
      else if (dragon[ai][aj].vitality < 10)
	return 2 * dragon_size;
      else
	return dragon_size;
    }
  }
    
  if (status1 == DEAD) {
    if (status2 == ALIVE)
      return 6 * dragon_size;
    if (2*dragon[ai][aj].genus + 2*dragon[bi][bj].genus
	+ dragon[ai][aj].heyes + dragon[bi][bj].heyes >=4)
      return 6 * dragon_size;
    if (status2 == UNKNOWN) {
      if (dragon[bi][bj].vitality < 0)
	return dragon_size;
      else if (dragon[bi][bj].vitality < 10)
	return 2 * dragon_size;
      else
	return 4 * dragon_size;
    }
  }

  return 0;
}

/*
 * Combine the reasons for a move at (m, n) into an old style value.
 * These heuristics are very ad hoc and not really expected to be very good.
 */
static int
value_move_reasons(int m, int n, int color)
{
  int k;
  int ai = -1;
  int aj = -1;
  int bi = -1;
  int bj = -1;
  int dragon1;
  int dragon2;
  
  /* Strategical value of connecting or cutting dragons. */
  static int dragon_value[MAX_DRAGONS];

  int cutstone = 0; /* Do we attack or defend at least one cutstone or
		       potential cutstone? */
  int tot_value = 0;
  int secondary_value = 0;
  for (k=0; k<MAX_DRAGONS; k++)
    dragon_value[k] = 0;
  
  for (k=0; k<MAX_REASONS; k++) {
    int this_value = 0;
    int r = move[m][n].reason[k];
    if(r == -1)
      break;
    
    switch (move_reasons[r].type) {
      case ATTACK_MOVE:
      case DEFEND_MOVE:
	ai = stringi[move_reasons[r].what];
	aj = stringj[move_reasons[r].what];
	
	/* Attack on our stones. */
	if (move_reasons[r].type == ATTACK_MOVE
	    && p[ai][aj] == color) {
	  secondary_value += 1;
	  continue;
	}
	
	/* Defense of enemy stones. */
	if (move_reasons[r].type == DEFEND_MOVE
	    && p[ai][aj] != color) {
	  secondary_value += 1;
	  continue;
	}
	
	/* Defenseless stone */
	if (worm[ai][aj].defendi == -1) {
	  secondary_value += worm[ai][aj].size;
	  continue;
	}
	
	if (worm[ai][aj].cutstone>0) {
	  cutstone = 1;
	  this_value = (2+2*worm[ai][aj].cutstone) *
	    (2+2*worm[ai][aj].size + worm[ai][aj].liberties);
	}
	else {
	  if (worm[m][n].liberties < 4 || dragon[m][n].status != DEAD)
	    this_value = 4*worm[ai][aj].size + 2*worm[ai][aj].liberties
	      + worm[m][n].liberties2;
	}
	if (worm[ai][aj].cutstone2 > 1)
	  this_value += 50 * (worm[ai][aj].cutstone2 - 1);
	tot_value += this_value;
	break;
	
      case CONNECT_MOVE:
      case CUT_MOVE:
	dragon1 = conn_dragon1[move_reasons[r].what];
	dragon2 = conn_dragon2[move_reasons[r].what];
	ai = dragoni[dragon1];
	aj = dragonj[dragon1];
	bi = dragoni[dragon2];
	bj = dragonj[dragon2];
	
	this_value = connection_value2(ai, aj, bi, bj);
	if (this_value > dragon_value[dragon1])
	  dragon_value[dragon1] = this_value;
	
	this_value = connection_value2(bi, bj, ai, aj);
	if (this_value > dragon_value[dragon2])
	  dragon_value[dragon2] = this_value;
	
	break;
	
      case ANTISUJI_MOVE:
	/* This move must not be played. End of story. */
	return 0;
    }
  }
  
  for (k=0; k<MAX_DRAGONS; k++)
    tot_value += dragon_value[k];
  
  if (cutstone)
    tot_value += 60;
  if (tot_value>0) {
    tot_value += (1+secondary_value)/2;
    tot_value += delta_moyo_simple(m, n, color);
  }
  
  return tot_value;
}

/*
 * Loop over all possible moves and value the move reasons for each.
 */
static void
value_moves(int color)
{
  int m;
  int n;
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++)
      move[m][n].value = value_move_reasons(m, n, color);
}

static int
experimental_attdef(int *i, int *j, int *val, int *equal_moves, int color)
{
  int m;
  int n;
  int tval;
  int found_one = 0;

  value_moves(color);
  
  for (m=0; m<board_size; m++)
    for (n=0; n<board_size; n++) {
      if (move[m][n].value == 0)
	continue;

      tval = move[m][n].value;
      TRACE("Experimental attacker/defender values %m to %d\n", 
	    m, n, tval);
      move_considered(m, n, tval);
      if (BETTER_MOVE(tval, *val, *equal_moves)) {
	*i = m;
	*j = n;
	found_one=1;
      }
    }
  return found_one;
}

#endif

/* 
 * Generate computer move for COLOR.
 *
 * Return the generated move in (*i, *j).
 */
  
int
genmove(int *i, int *j, int color)
{
  int val;
  int equal_moves;
  int m,n;
  int shapei, shapej;

  /* prepare our table of moves considered */
  memset(potential_moves, 0, sizeof(potential_moves));

  /* prepare our table of move reasons */
  clear_move_reasons();

  /* Reset all the statistics for each move. */
  stats.nodes = 0;

  /* prepare matchpat for use */
  compile_for_match();

#if HASHING
  /* Initialize things for hashing of positions. */
  hashtable_clear(movehash);
  board_to_position(p, ko_i, ko_j, /*color,*/ &(hashdata.hashpos));
  hashdata.hashval = board_hash(p, ko_i, ko_j/*, color*/);
#endif

  /* no move is found yet. */
  *i = -1;  
  *j = -1;  
  val = -1; 

  /* Used to give every equal move the same chance to be chosen. */
  equal_moves = 2;

  {
    /* Don't print reading traces during make_worms and make_dragons unless 
       the user really wants it (verbose == 3). */

    int save_verbose=verbose;
    if ((verbose==1) || (verbose==2)) --verbose;
    if (make_worms()) {
      make_dragons();
    }
    verbose=save_verbose;
  }

  /* Map the moyos for both sides. */
  make_moyo(color);
  
  if (printworms) {
    show_dragons();
  }
  else if (verbose) {
    for (m=0;m<board_size;m++)
      for (n=0;n<board_size;n++)
	if ((p[m][n])
	    && (worm[m][n].origini==m) 
	    && (worm[m][n].originj==n) 
	    && (worm[m][n].cutstone)) 
	{
	  TRACE("%scutting stones found at %m\n",
		worm[m][n].cutstone==2 ? "potential " : "", m,n); 
	}
  }	  
  
  if (printmoyo) {
    print_moyo(color);
  }
   
  if (printboard)
    showboard(printboard);
  
  assert(stackp == 0);  /* stack empty */

  /* Try to find nice fuseki moves. */
  if (fuseki(i, j, &val, &equal_moves, color))
  {
      TRACE("Fuseki Player likes %m with value %d\n", *i, *j, val);
      if(analyzerflag&ANALYZE_RECOMMENDED)
          sgfBoardText(0,*i,*j,"F");
  }
  assert(stackp == 0);  /* stack empty */

  /* Try to catch your stones in semeai. */
  if (semeai(i, j, &val, &equal_moves, color))
  {
     TRACE("Semeai Player likes %m with value %d\n", *i, *j, val);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"S");
  }
  assert(stackp == 0);  /* stack empty */
  
  /* pattern matcher */
  if (shapes(i, j, &val, &equal_moves, color))
  {
     TRACE("Shape Seer likes %m with value %d\n", *i, *j, val);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"P");
  }
  shapei = *i;
  shapej = *j;

  assert(stackp == 0);  /* stack empty */

#if EXPERIMENTAL_MOVEGEN

  induce_secondary_move_reasons(color);
  
  /* List reasons for a move. Later this will be used to replace attacker(),
   * defender(), and eyefinder().
   */
  list_move_reasons(color);
  
  if (experimental_attdef(i, j, &val, &equal_moves, color))
  {
     TRACE("Experimental attacker/defender likes %m with value %d\n",
	   *i, *j, val);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"A");
  }
  assert(stackp == 0);  /* stack empty */
 
#else
  
  /* Try to catch your stones in a regular fight (semeai already done above). */
  if (attacker(i, j, &val, &equal_moves, color, shapei, shapej))
  {
     TRACE("Attacker likes %m with value %d\n", *i, *j, val);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"A");
  }
  assert(stackp == 0);  /* stack empty */
  
  /* Try to save my stones if you can attack them. */
  if (defender(i, j, &val, &equal_moves, color, shapei, shapej))
  {  TRACE("Defender likes %m with value %d\n", *i, *j, val);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"D");
  }
  assert(stackp == 0);  /* stack empty */

#endif
  
  /* Look for eye stealing moves. */
  if (eye_finder(i, j, &val, &equal_moves, color, shapei, shapej))
  {  TRACE("Eye Finder  likes %m with value %d\n", *i, *j, val);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"E");
  }
  assert(stackp == 0);  /* stack empty */
  
  /* If no move found yet, revisit any semeai and change the
   * status of the opponent group from DEAD to UNKNOWN, then 
   * run shapes again. This may turn up a move.
   */
  if ((val < 0) 
      && revise_semeai(color) 
      && shapes(i, j, &val, &equal_moves, color))
    {
       TRACE("Upon reconsideration Shape Seer likes %m with value %d\n",
	  *i, *j, val); 
       if(analyzerflag&ANALYZE_RECOMMENDED)
         sgfBoardText(0,*i,*j,"p");
    }
  if ((val < 0) && fill_liberty(i, j, &val, color))
  {
     TRACE("Filling a liberty at %m\n", *i, *j);
     if(analyzerflag&ANALYZE_RECOMMENDED)
       sgfBoardText(0,*i,*j,"L");
  }
  if (style & STY_FEARLESS) {

    /* If there is a big move
          and the temperature is low (i.e. no fighting)
          and the recommended move is somewhere near the big move
       then do the big move instead. */
    if (very_big_move[0] 
	&& val < min(75, very_big_move[0]) 
	&& (max(abs(*i - very_big_move[1]), 
		abs(*j - very_big_move[2])) >= 3)) 
    {
      if (printmoyo & 128)
        gprintf("genmove() recommended %m val %d ", *i, *j, val);
      else
        TRACE("genmove() recommended %m val %d ", *i, *j, val);

      *i = very_big_move[1];
      *j = very_big_move[2];
      val = very_big_move[0];
      if (printmoyo & 128)
        gprintf("but we play %m with new val %d, t %d\n",
		*i, *j, val,terri_eval[0]);
      else
        TRACE("but we play %m with new val %d, t %d\n",
	      *i, *j, val,terri_eval[0]);
      if(analyzerflag&ANALYZE_RECOMMENDED)
        sgfBoardText(0,*i,*j,"B");
    }
  }

  /* If no move is found then pass. */
  if (val < 0) {
    TRACE("I pass.\n");
    *i = board_size;
    *j = board_size;
  } else
    TRACE("genmove() recommends %m with value %d\n", *i, *j, val);
  
  /* If statistics is turned on, this is the place to show it. */
  if (showstatistics) {
    char text[120];
    sprintf(text,"Nodes: %d\n", stats.nodes);
    sgfAddComment(0,text);
    gprintf("Nodes: %d\n", stats.nodes);
      
  }

  return val;
  
}  /* end genmove */




/* This is called for each move which has been considered. For
 * debugging purposes, we keep a table of all the moves we
 * have considered.
 */

void 
move_considered(int i, int j, int val)
{
  if ((val > potential_moves[i][j])) {
    potential_moves[i][j] = val;
  }
}



/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */

